/*
 * Licensed Materials - Property of IBM                               
 * Component IBM Rational ROSE_RT BUILD : TARGET_RTS                  
 * Copyright IBM Corporation 1999, 2008. All Rights Reserved.       
 * U.S. Government Users Restricted Rights -  Use, duplication or     
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

#include <RTActor_class.h>
#include <RTDaemon.h>
#include <RTDaemonInfo.h>
#include <RTDebugger.h>
#include <RTDiag.h>
#include <RTDiagStream.h>
#include <RTElasticArray.h>
#include <RTEventInfo.h>
#include <RTFormat.h>
#include <RTMemoryUtil.h>
#include <RTMessage.h>
#include <RTPortDescriptor.h>
#include <RTProtocol.h>

static int getStateDepth( const RTActor_class & info,
			  int                   theState,
			  int                   cpParent )
{
    int depth  = 0;
    int state  = theState;
    int parent = theState <= info.numStates
			   ? info.parentState[ theState - 1 ]
			   : cpParent;

    while( parent > 0 )
    {
	++depth;
	state  = parent;
	parent = info.parentState[ state - 1 ];
    }

    return depth;
}

static char * getStateStr( const RTActor_class & info,
			   int                   theState,
			   int                   cpParent )
{
    // return a dynamically allocated string with the state 'path'

    // no state information, return an empty string
    if( theState == 0 )
	return RTMemoryUtil::strdup( "" );

    int parent = theState <= info.numStates
			   ? info.parentState[ theState - 1 ]
			   : cpParent;

    // the top state is just a single colon (':')
    if( parent == 0 )
	return RTMemoryUtil::strdup( ":" );

    int length = 0;
    int state  = theState;

    while( parent > 0 )
    {
	length += 1 + RTMemoryUtil::strlen( info.names[ state - 1 ] );

	state  = parent;
	parent = info.parentState[ state - 1 ];
    }

    char * returnStr = new char[ (unsigned)length + 1U ];
    char * pointer   = &returnStr[ length ];

    *pointer = '\0';
    state    = theState;
    parent   = state <= info.numStates
		      ? info.parentState[ state - 1 ]
		      : cpParent;

    while( parent > 0 )
    {
	length   = RTMemoryUtil::strlen( info.names[ state - 1 ] );
	pointer -= length;

	RTMemoryUtil::memcpy( pointer, info.names[ state - 1 ], length );
	*--pointer = ':';

	state  = parent;
	parent = info.parentState[ state - 1 ];
    }

    return returnStr;
}

static const RTPortDescriptor * getPort( const RTActor_class & info, int id )
{
    const RTPortDescriptor * port = info.ports;

    for( int count = info.numPorts; --count >= 0; ++port )
	if( port->id == id )
	    return port;

    return (const RTPortDescriptor *)0;
}

static char * getPortNameId( const RTActor_class & info, int portId, int index )
{
    const RTPortDescriptor * port = getPort( info, portId );

    if( port == (const RTPortDescriptor *)0 )
	return (char *)0;
    else
    {
	int    portLength = RTMemoryUtil::strlen( port->name );
	char * tempPort   = new char[ (unsigned)portLength + 15U ];

	if( tempPort != (char *)0 )
	{
	    RTMemoryUtil::memcpy( tempPort, port->name, portLength );
	    tempPort[ portLength ] = ':';
	    (void)RTFormat::_int( index, &tempPort[ portLength + 1 ], 14 );
	}

	return tempPort;
    }
}

static char * getPortNameMsg( const RTActor_class & info,
			      const RTMessage     * msg )
{
    RTProtocol * port = msg->sap();

    if( port == (RTProtocol *)0 )
	return RTMemoryUtil::strdup( "<system>" );
    else
	return getPortNameId( info, port->getId(), msg->sapIndex0() );
}

static bool evtChainMatches( RTEventInfo  & eventInfo,
			     RTDaemonInfo * daemonInfo )
{
    // create the event origin string, if it doesn't exist already
    if( eventInfo.chainStrOrigin == (char *)0 )
    {
	eventInfo.chainStrOrigin = getStateStr( *daemonInfo->info,
						eventInfo.chainOrigin,
						eventInfo.chainParent );
    }

    // create the event port string, if it doesn't exist already
    if( eventInfo.portName == (char *)0 && eventInfo.message != (RTMessage *)0 )
    {
	eventInfo.portName = getPortNameMsg( *daemonInfo->info,
					     eventInfo.message );
    }

    if( daemonInfo->daemonKind != RTDaemonInfo::InfoActor &&
	daemonInfo->daemonKind != RTDaemonInfo::InfoChain )
    {
	return false;
    }

    // false if daemonInfo.chainOrigin is set to something else
    if( daemonInfo->chainOrigin >= 0 &&
	daemonInfo->chainOrigin != eventInfo.chainOrigin )
    {
	return false;
    }

    // false if daemonInfo.chainPosition is set to something else
    if( daemonInfo->chainPosition >= 0 &&
	daemonInfo->chainPosition != eventInfo.chainPosition )
    {
	return false;
    }

    // false if the origin strings don't match
    if( daemonInfo->chainStrOrigin != (char *)0 &&
	RTMemoryUtil::strcmp( daemonInfo->chainStrOrigin,
			      eventInfo.chainStrOrigin ) != 0 )
    {
	return false;
    }

    // false if the trans strings don't match
    if( daemonInfo->chainStrTrans != (char *)0 &&
	RTMemoryUtil::strcmp( daemonInfo->chainStrTrans,
			      eventInfo.chainStrTrans ) != 0 )
    {
	return false;
    }

    if( daemonInfo->daemonKind == RTDaemonInfo::InfoChain )
	daemonInfo->chainOrigin = eventInfo.chainOrigin;

    return true;
}

static bool evtStateMatches( RTEventInfo        & eventInfo,
			     const RTDaemonInfo * daemonInfo )
{
    // create the event target string, if it doesn't exist already
    if( eventInfo.stateStrTarget == (char *)0 )
    {
	eventInfo.stateStrTarget = getStateStr( *daemonInfo->info,
						eventInfo.stateTarget,
						eventInfo.chainParent );
    }

    // create the event port string, if it doesn't exist already
    if( eventInfo.portName == (char *)0 )
    {
	eventInfo.portName = getPortNameMsg( *daemonInfo->info,
					     eventInfo.message );
    }

    // false if not a state
    if( daemonInfo->daemonKind != RTDaemonInfo::InfoState )
	return false;

    // false if the target strings don't match
    if( daemonInfo->stateStrTarget != (char *)0 &&
	RTMemoryUtil::strcmp( daemonInfo->stateStrTarget,
			      eventInfo  .stateStrTarget ) != 0 )
    {
	return false;
    }

    return true;
}

static bool evtFSMMatches( RTEventInfo        & eventInfo,
			   const RTDaemonInfo * daemonInfo )
{
    // create the event origin string, if it doesn't exist already
    if( eventInfo.chainStrOrigin == (char *)0 )
    {
	eventInfo.chainStrOrigin = getStateStr( *daemonInfo->info,
						eventInfo.chainOrigin,
						eventInfo.chainParent );
    }

    if( daemonInfo->daemonKind == RTDaemonInfo::InfoState )
    {
	int strTargetLen = RTMemoryUtil::strlen( daemonInfo->stateStrTarget );

	// event strOrigin should contain the daemon strTarget
	if( RTMemoryUtil::strcmp( daemonInfo->stateStrTarget,
				  eventInfo.chainStrOrigin,
				  strTargetLen ) != 0 )
	{
	    return false;
	}

	// event state depth should not be > daemon state depth + 1
	int eventDepth = getStateDepth(  *daemonInfo->info,
					  eventInfo.chainOrigin,
					  eventInfo.chainParent );

	if( eventDepth > daemonInfo->stateDepth + 1 )
	    return false;
    }
    // InfoActor: anything matches
    else if( daemonInfo->daemonKind != RTDaemonInfo::InfoActor )
	return false;

    return true;
}

static bool evtMsgInMatches( RTEventInfo        & eventInfo,
			     const RTDaemonInfo * daemonInfo )
{
    // create the event port string, if it doesn't exist already
    if( eventInfo.portName == (char *)0 )
    {
	eventInfo.portName = getPortNameMsg( *daemonInfo->info,
					     eventInfo.message );
    }

    // false if not a port
    if( daemonInfo->daemonKind != RTDaemonInfo::InfoActor &&
	daemonInfo->daemonKind != RTDaemonInfo::InfoPort )
    {
	return false;
    }

    // false unless port ids match
    if( daemonInfo->portId != eventInfo.message->sapId() )
	return false;

    // false unless port ref or port indexes match
    if( daemonInfo->portIndex != -1 &&
	daemonInfo->portIndex != eventInfo.message->sapIndex0() )
    {
	return false;
    }

    return true;
}

static bool evtMsgOutMatches( RTEventInfo        & eventInfo,
			      const RTDaemonInfo * daemonInfo )
{
    // create the event port string, if it doesn't exist already
    if( eventInfo.portName == (char *)0 )
    {
	eventInfo.portName = getPortNameId( *daemonInfo->info,
					    eventInfo.message->_sendPortId,
					    eventInfo.message->_sendPortIndex );
    }

    // false if not a port
    if( daemonInfo->daemonKind != RTDaemonInfo::InfoPort )
	return false;

    // false unless port ids match
    if( daemonInfo->portId != eventInfo.message->_sendPortId )
	return false;

    // false unless port ref or port indexes match
    if( daemonInfo->portIndex != -1 &&
	daemonInfo->portIndex != eventInfo.message->_sendPortIndex )
    {
	return false;
    }

    // Update eventInfo.info because it will be used in addEvent
    eventInfo.info = daemonInfo->info;

    return true;
}

static bool evtMsgOutInMatches( RTEventInfo        & eventInfo,
				const RTDaemonInfo * daemonInfo )
{
    // create the event port string, if it doesn't exist already
    if( eventInfo.portName == (char      *)0 &&
	eventInfo.message  != (RTMessage *)0 )
    {
	eventInfo.portName = getPortNameMsg( *daemonInfo->info,
					     eventInfo.message );
    }

    // false if event probe is 'any probe' or invalid
    if( eventInfo.probeNr < 0 )
	return false;

    // false if not an actor
    if( daemonInfo->daemonKind != RTDaemonInfo::InfoActor )
	return false;

    return true;
}

static bool evtVariableMatches( const RTEventInfo  & eventInfo,
				const RTDaemonInfo * daemonInfo )
{
    // false if not a variable
    if( daemonInfo->daemonKind != RTDaemonInfo::InfoVariable )
	return false;

    // false if variable names don't match
    if( RTMemoryUtil::strcmp( daemonInfo->variableName,
			      eventInfo.varName ) != 0 )
	return false;

    return true;
}

static bool evtHistoryMatches( RTEventInfo        & eventInfo,
			       const RTDaemonInfo * daemonInfo )
{
    // false if not a state
    if( daemonInfo->daemonKind != RTDaemonInfo::InfoState )
	return false;

    // false if different states
    if( daemonInfo->stateTarget != eventInfo.stateTarget )
	return false;

    if( eventInfo.histStrHist == (char *)0 )
	eventInfo.histStrHist = getStateStr( *daemonInfo->info,
					     eventInfo.histHist, 1 );

    if( eventInfo.histStrCurr == (char *)0 )
	eventInfo.histStrCurr = getStateStr( *daemonInfo->info,
					     eventInfo.histCurr, 1 );

    return true;
}

RTDaemonInfo * RTDaemon::eventMatches( RTEventInfo & eventInfo )
{
    for( int iActor = 0; iActor < actorInfo.size(); ++iActor )
    {
	RTDaemonInfo * daemonInfo = (RTDaemonInfo *)actorInfo[ iActor ];

	if( daemonInfo == (RTDaemonInfo *)0 )
	    continue;

	// skip unless event says any probe or probeNumbers match
	if( eventInfo.probeNr >= 0 &&
	    eventInfo.probeNr != daemonInfo->probeNr )
	{
	    continue;
	}

	// skip unless daemon contains actor ref or actorIndexes match
	if( eventInfo.actorIndex   >= 0 &&
	    daemonInfo->actorIndex >= 0 &&
	    daemonInfo->actorIndex != eventInfo.actorIndex )
	{
	    continue;
	}

	if( eventInfo.eventKind == RTEventInfo::EventPresence )
	    return daemonInfo; // all daemons want this info

	else if( eventInfo.eventKind == RTEventInfo::EventControl )
	    return daemonInfo; // all daemons want this info

	else if( daemonInfo->info == (const RTActor_class *)0 )
	    continue;

	switch( eventInfo.eventKind )
	{
	case RTEventInfo::EventReference:
	    if( daemonInfo->daemonKind == RTDaemonInfo::InfoActor )
		return daemonInfo;
	    break;

	case RTEventInfo::EventChain:
	    if( evtChainMatches( eventInfo, daemonInfo ) )
		return daemonInfo;
	    break;

	case RTEventInfo::EventState:
	    if( evtStateMatches( eventInfo, daemonInfo ) )
		return daemonInfo;
	    break;

	case RTEventInfo::EventFSM:
	case RTEventInfo::EventLateFSM:
	    if( evtFSMMatches( eventInfo, daemonInfo ) )
		return daemonInfo;
	    break;

	case RTEventInfo::EventMsgIn:
	    if( evtMsgInMatches( eventInfo, daemonInfo ) )
		return daemonInfo;
	    break;

	case RTEventInfo::EventMsgOut:
	    if( evtMsgOutMatches( eventInfo, daemonInfo ) )
		return daemonInfo;
	    break;

	case RTEventInfo::EventMsgOutIn:
	case RTEventInfo::EventTimer:
	    if( evtMsgOutInMatches( eventInfo, daemonInfo ) )
		return daemonInfo;
	    break;

	case RTEventInfo::EventVariable:
	    if( evtVariableMatches( eventInfo, daemonInfo ) )
		return daemonInfo;
	    break;

	case RTEventInfo::EventHistory:
	    if( evtHistoryMatches( eventInfo, daemonInfo ) )
		return daemonInfo;
	    break;

#ifdef _lint
	case RTEventInfo::EventPresence:
	case RTEventInfo::EventControl:
#endif
	default:
	    break;
	}
    }

    return (RTDaemonInfo *)0;
}
