/*
 * Licensed Materials - Property of IBM                               
 * Component IBM Rational ROSE_RT BUILD : TARGET_RTS                  
 * Copyright IBM Corporation 1999, 2008. All Rights Reserved.       
 * U.S. Government Users Restricted Rights -  Use, duplication or     
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

#include <RTMemoryUtil.h>
#include <RTProtocolAdapter.h>
#include <RTProtocolDescriptor.h>
#include <RTSignalDescriptor.h>

// Build a signal map which when indexed by (outSignal - 1) from tx,
// yields the appropriate inSignal from rx. Invalid signals are mapped
// to zero. The algorithm assumes the signals are sorted in ascending
// order according to their names.

static void buildMap( const RTProtocolDescriptor * tx,
		      const RTProtocolDescriptor * rx,
		      int                        * map )
{
    const RTProtocolDescriptor * conjugate = tx->conjugate;
    const RTSignalDescriptor   * txSignal  = conjugate->inSignals;
    int                          txCount   = conjugate->numInSignals;
    const RTSignalDescriptor   * rxSignal  = rx->inSignals;
    int                          rxCount   = rx->numInSignals;

    --map; // signal values start at 1

    for( ; --txCount >= 0; ++txSignal, ++rxSignal )
    {
	int rxValue;

	for(;;)
	{
	    if( rxCount <= 0 )
	    {
		// no more signals in rx
		rxValue = 0;
		break;
	    }

	    int cmp = RTMemoryUtil::strcmp( txSignal->name, rxSignal->name );

	    if( cmp < 0 )
	    {
		// signals are sorted, the one we want should appear
		// before this one: it must not be there
		rxValue = 0;
		break;
	    }

	    if( cmp == 0 )
	    {
		rxValue = rxSignal->value;
		break;
	    }
	    ++rxSignal;
	    --rxCount;
	}

	map[ txSignal->value ] = rxValue;
    }
}

//lint -save -e946

void RTProtocolAdapter::adapt( const RTProtocolDescriptor  * source,
			       const RTProtocolDescriptor  * target,
			       const int                  ** sourceMap,
			       const int                  ** targetMap )
{
    if( source->conjugate == target )
    {
	*sourceMap = (const int *)0;
	*targetMap = (const int *)0;
	return;
    }

    // We avoid duplication by only storing entries with source <= target.
    if( source > target )
    {
	adapt( target, source, targetMap, sourceMap );
	return;
    }

    const int sourceNumOutSignals = source->conjugate->numInSignals;
    const int targetNumOutSignals = target->conjugate->numInSignals;

    // Do a binary search for the desired entry.
    int lo = 0;
    int hi = used;

    while( lo < hi )
    {
	int           mid = ( lo + hi ) / 2;
	const Entry * ent = entry[ mid ];

	if     ( source < ent->source ) hi = mid;
	else if( source > ent->source ) lo = mid + 1;
	else if( target < ent->target ) hi = mid;
	else if( target > ent->target ) lo = mid + 1;
	else
	{
	    *sourceMap = &ent->maps[ 0 ];
	    *targetMap = &ent->maps[ sourceNumOutSignals ];
	    return;
	}
    }

    // Grow the table if necessary.
    if( used == size )
    {
	size += 64;

	Entry ** new_entry = new Entry * [ (unsigned)size ];

	for( int i = used; --i >= 0; )
	    new_entry[ i ] = entry[ i ];

	delete [] entry;

	entry = new_entry;
    }

    // Make space for new entry.

    for( int i = used; --i >= hi; )
	entry[ i + 1 ] = entry[ i ];

    int words = (int)( ( sizeof( Entry ) + sizeof( int ) - 1U )
		       / sizeof( int ) )
	      + sourceNumOutSignals
	      + targetNumOutSignals
	      - 1;

    Entry * new_entry = (Entry *)(void *)new int [ (unsigned)words ];

    new_entry->source = source;
    new_entry->target = target;

    *sourceMap = &new_entry->maps[ 0 ];
    buildMap( source, target, &new_entry->maps[ 0 ] );

    *targetMap = &new_entry->maps[ sourceNumOutSignals ];
    buildMap( target, source, &new_entry->maps[ sourceNumOutSignals ] );

    entry[ hi ] = new_entry;

    used += 1;
}

//lint -restore
